------------------------------------------------------------------------
-- Implementation of a splay heap.
--
-- Based on Okasaki's Purely Functional Data Structures and McBride's
-- How to Keep Your Neighbours in Order.
------------------------------------------------------------------------

{-# OPTIONS --safe --without-K #-}

open import Relation.Binary.Bundles

module SplayHeap {c ℓ₁ ℓ₂} (T : TotalOrder c ℓ₁ ℓ₂) where

open TotalOrder T hiding (refl)

open import Data.Nat.Base using (ℕ; zero; suc; _+_)
open import Data.Nat.Properties using (n<1+n; +-suc; +-assoc)
open import Data.Product.Base
open import Data.Sum.Base using (inj₁; inj₂)
open import Function.Base
open import Level using (_⊔_)
open import Relation.Binary.Construct.Add.Extrema.NonStrict _≤_
open import Relation.Binary.PropositionalEquality using (_≡_; refl; cong; module ≡-Reasoning)
open import Relation.Nullary.Construct.Add.Extrema
open import Relation.Nullary.Decidable.Core

data Tree (l u : Carrier ±) : Set (c ⊔ ℓ₂) where
  leaf : .(l ≤± u) → Tree l u
  node : (x : Carrier) → Tree l [ x ] → Tree [ x ] u → Tree l u

Heap : Set (c ⊔ ℓ₂)
Heap = Tree ⊥± ⊤±

data NonEmpty {l u} : Tree l u → Set (c ⊔ ℓ₂) where
  instance nonEmpty : ∀ {x l r} → NonEmpty (node x l r)

nonEmpty? : ∀ {l u} (t : Tree l u) → Dec (NonEmpty t)
nonEmpty? (leaf _) = no λ ()
nonEmpty? (node _ _ _) = yes nonEmpty

private
  -- linear in length of spine :(
  -- It seems to compile away fine, though.
  slacken⊥ : ∀ {l l′ u} → Tree l u → .(l′ ≤± l) → Tree l′ u
  slacken⊥ (leaf l≤u) l′≤l = leaf (≤±-trans trans l′≤l l≤u)
  slacken⊥ (node x a b) l≤l′ = node x (slacken⊥ a l≤l′) b

  partition : ∀ {l u} (p : Carrier) → .(l ≤± [ p ]) → .([ p ] ≤± u) → Tree l u → Tree l [ p ] × Tree [ p ] u
  partition p l≤p p≤u (leaf _) = leaf l≤p , leaf p≤u
  partition p l≤p p≤u (node x a b) with total x p
  partition p l≤p p≤u (node x a (leaf _)) | inj₁ x≤p = node x a (leaf [ x≤p ]) , leaf p≤u
  partition p l≤p p≤u (node x a (node y b c)) | inj₁ x≤p with total y p
  ... | inj₁ y≤p with small , big ← partition p [ y≤p ] p≤u c = node y (node x a b) small , big
  ... | inj₂ p≤y with small , big ← partition p [ x≤p ] [ p≤y ] b = node x a small , node y big c
  partition p l≤p p≤u (node x (leaf _) a) | inj₂ p≤x = leaf l≤p , node x (leaf [ p≤x ]) a
  partition p l≤p p≤u (node x (node y a b) c) | inj₂ p≤x with total y p
  ... | inj₁ y≤p with small , big ← partition p [ y≤p ] [ p≤x ] b = node y a small , node x big c
  ... | inj₂ p≤y with small , big ← partition p l≤p [ p≤y ] a = small , (node y big (node x b c))

insert : Carrier → Heap → Heap
insert x t with l , r ← partition x ⊥±≤[ x ] [ x ]≤⊤± t = node x l r

findDeleteMin : ∀ {l u} (t : Tree l u) → .{{NonEmpty t}} → Carrier × Tree l u
findDeleteMin (node x (leaf l≤x) a) = x , slacken⊥ a l≤x
findDeleteMin (node x (node y (leaf l≤y) a) b) = y , node x (slacken⊥ a l≤y) b
findDeleteMin (node x (node y a@(node _ _ _) b) c) with z , a′ ← findDeleteMin a = z , node y a′ (node x b c)

empty : Heap
empty = leaf ⊥±≤⊤±

-- number of elements in the heap
-- useful for induction proofs hopefully
#_ : ∀ {l u} → Tree l u → ℕ
# leaf _ = 0
# node x a b = suc (# a + # b)

private
  #-slacken⊥ : ∀ {l l′ u} (t : Tree l u) .(l′≤l : l′ ≤± l) → # slacken⊥ t l′≤l ≡ # t
  #-slacken⊥ (leaf x) l′≤l = refl
  #-slacken⊥ (node x a b) l′≤l = cong (suc ∘ (_+ # b)) (#-slacken⊥ a l′≤l)

#-findDeleteMin : ∀ {l u} (t : Tree l u) → .{{_ : NonEmpty t}} → suc (# proj₂ (findDeleteMin t)) ≡ # t
#-findDeleteMin (node x (leaf l≤x) b) = cong suc (#-slacken⊥ b l≤x)
#-findDeleteMin (node x (node y (leaf l≤y) b) c) = cong (suc ∘ suc ∘ (_+ # c)) (#-slacken⊥ b l≤y)
#-findDeleteMin (node x (node y a@(node _ s t) b) c) = cong (suc ∘ suc) $ begin
  # proj₂ (findDeleteMin a) + suc (# b + # c)   ≡⟨ +-suc (# proj₂ (findDeleteMin a)) (# b + # c) ⟩
  suc (# proj₂ (findDeleteMin a)) + (# b + # c) ≡⟨ cong (_+ (# b + # c)) (#-findDeleteMin a) ⟩
  suc (# s + # t) + (# b + # c)                 ≡˘⟨ cong suc (+-assoc (# s + # t) (# b) (# c)) ⟩
  suc (# s + # t + # b + # c)                   ∎
  where open ≡-Reasoning
